/*
 * Copyright (c) 1999, 2013, Oracle and/or its affiliates. All rights reserved.
 * ORACLE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

import java.io.FileInputStream;
import java.io.IOException;
import java.net.InetAddress;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import sun.security.krb5.KrbException;

/**
 * Helper class to start local KDC server. The class does the following:
 * - parses command line options (host, port, realm, principals list, etc)
 * - loads Kerberos principals list
 * - creates Kerberos configuration to file
 * - created keytab file
 *
 * @author Artem Smotrakov
 */
public class KDCLauncher {

    private Map<String, String> principals;
    private String realm;
    private String ktab;
    private String conf;
    private String host;
    private int port;
    private KDC kdc;
    private boolean daemon = false;

    KDCLauncher(String realm, String ktab, String conf, String host, int port, Map<String, String> principals) {
        this.realm = realm;
        this.ktab = ktab;
        this.conf = conf;
        this.host = host;
        this.port = port;
        this.principals = principals;
    }

    /**
     * Start KDC server on specified port:
     * - create KDC instance (the method throws IllegalStateException if KDC has already started)
     * - create Kerberos principals
     * - save Kerberos configuration
     * - save keytab file
     * - no pre-auth required
     */
    public void start() throws IOException, KrbException {
        if(kdc != null) {
            throw new IllegalStateException("KDC has already started");
        }

        if(host == null) {
            InetAddress localAddress = InetAddress.getLocalHost();
            host = localAddress.getCanonicalHostName();
        }

        System.out.println("Start KDC on " + host + ":" + port + " port with " + realm + " realm");
        kdc = KDC.create(realm, host, port, daemon);
        kdc.setOption(KDC.Option.PREAUTH_REQUIRED, Boolean.FALSE);

        if(principals != null) {
            for(Map.Entry<String, String> entry : principals.entrySet()) {
                String principalName = entry.getKey();
                String principalPassword = entry.getValue();
                if(principalPassword == null || principalPassword.isEmpty()) {
                    System.out.println("Add principal '" + principalName + "' with random password");
                    kdc.addPrincipalRandKey(principalName);
                } else {
                    System.out.println("Add principal '" + principalName + "' with '" + principalPassword + "' password");
                    kdc.addPrincipal(principalName, principalPassword.toCharArray());
                }
            }
        }

        if(conf != null) {
            System.out.println("Save configuration file to " + conf);
            KDC.saveConfig(conf, kdc, "forwardable = true");
        }

        if(ktab != null) {
            System.out.println("Save keytab file to " + ktab);
            kdc.writeKtab(ktab, false);
        }
    }

    /**
     * Entry point:
     * - parse command line options
     * - start KDC server
     */
    public static void main(String[] args) throws IOException, KrbException {
        KDCLauncher kdcLauncher = KDCLauncher.init(args);
        kdcLauncher.start();
    }

    /**
     * Parse command line options and create KDC instance.
     *
     * Command line options:
     * -principals: path to file that contains list of Kerberos principals
     * -realm: Kerberos realm
     * -host: host name
     * -port: port number for KDC server
     * -ktab: path to file where keytab data is saved
     * -conf: path to file where Kerberos configuration is saved
     *
     * @param args Command line options
     * @return KDC instance
     */
    private static KDCLauncher init(String[] args) {
        Map<String, String> principals = null;
        String realm = null;
        String ktab = null;
        String conf = null;
        String host = null;
        int port = 0;

        try {
            int i = 0;
            String arg;
            while (i < args.length) {
                arg = args[i++];
                switch (arg) {
                    case "-principals":
                        if (i < args.length)
                            principals = loadPrincipals(args[i++]);
                        else
                            throw new Exception("-principals requires a filename");
                        break;
                    case "-realm":
                        if (i < args.length)
                            realm = args[i++];
                        else
                            throw new Exception("-realm requires realm name");
                        break;
                    case "-host":
                        if (i < args.length)
                            host = args[i++];
                        else
                            throw new Exception("-host requires host name");
                        break;
                    case "-port":
                        if (i < args.length)
                            port = Integer.parseInt(args[i++]);
                        else
                            throw new Exception("-port requires a port number");
                        break;
                    case "-ktab":
                        if (i < args.length)
                            ktab = args[i++];
                        else
                            throw new Exception("-ktab requires a filename");
                        break;
                    case "-conf":
                        if (i < args.length)
                            conf = args[i++];
                        else
                            throw new Exception("-conf requires a filename");
                        break;
                    default:
                        throw new Exception("Unrecognized option: " + arg);
                }
            }
        } catch(Exception e) {
            System.out.println("Parsing command line options failed: " + e.getMessage());
            System.out.println("Usage: " + KDCLauncher.class.getName() + " -realm REALM.NAME -port 1234 -ktab path/to/ktab -conf path/to/conf -principals path/to/principals");
            throw new IllegalArgumentException("Wrong parameters");
        }

        if(realm == null || realm.isEmpty()) {
            System.out.println("Realm is not set. Use -realm parameter");
            throw new IllegalArgumentException("Wrong parameters");
        }

        if(port <= 0 || port >= 65535) {
            System.out.println("Wrong port number, or port is not set. Use -port parameter");
            throw new IllegalArgumentException("Wrong parameters");
        }

        return new KDCLauncher(realm, ktab, conf, host, port, principals);
    }

    /**
     * Load Kerberos principal list.
     */
    private static Map<String, String> loadPrincipals(String filename) throws IOException {
        Properties db = new Properties();
        db.load(new FileInputStream(filename));
        Map<String, String> principals = new HashMap<>();
        for(String principalName : db.stringPropertyNames()) {
            String principalPassword = db.getProperty(principalName);
            principals.put(principalName, principalPassword);
        }
        return principals;
    }
    
    void setDaemon(boolean daemon) {
        this.daemon = daemon;
    }
}
