import javax.net.ssl.SSLSocket;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLSocketFactory;
import java.io.IOException;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.logging.Logger;
import java.util.logging.Level;
import java.util.Arrays;
import java.util.List;
import javax.net.ssl.SNIHostName;
import javax.net.ssl.SNIServerName;
import javax.net.ssl.SSLParameters;

class SSLClient {

    private final static Logger logger = Logger.getLogger(SSLClient.class.getName());

    private SSLSocket socket;
    private DataExchanger dataExchanger;

    public static void main(String[] args) throws Exception {
        SSLClient sslClient = SSLClient.init(args);
        sslClient.connect();
    }

    static SSLClient init(String[] args) throws Exception {
        int port = 0;
        String host = null;
        String cipherSuiteFilter = null;
        String sniHostName = null;

        try {
            int i = 0;
            String arg;
            while (i < args.length) {
                arg = args[i++];
                switch (arg) {
                    case "-host":
                        if (i < args.length)
                            host = args[i++];
                        else
                            throw new Exception("-host requires a hostname");
                        break;
                    case "-port":
                        if (i < args.length)
                            port = Integer.parseInt(args[i++]);
                        else
                            throw new Exception("-port requires a port number");
                        break;
                    case "-cipher_suite_filter":
                        if (i < args.length)
                            cipherSuiteFilter = args[i++];
                        else
                            throw new Exception("-cipher_suite_filter requires a value");
                        break;
                    case "-sni_host_name":
                        if (i < args.length)
                            sniHostName = args[i++];
                        else
                            throw new Exception("-sni_host_name requires a value");
                        break;
                    default:
                        throw new Exception("Unrecognized option: " + arg);
                }
            }
        } catch(Exception e) {
            System.out.println("Parsing command line options failed: " + e.getMessage());
            System.out.println("Usage: " + SSLClient.class.getName() + " -port 1234");
            throw new IllegalArgumentException("Wrong parameters");
        }

        if(host == null || host.isEmpty()) {
            System.out.println("Wrong hostname, or hostname is not set. Use -host parameter");
            throw new IllegalArgumentException("Wrong parameters");
        }

        if(port <=0 || port > 65535) {
            System.out.println("Wrong port number, or port is not set. Use -port parameter");
            throw new IllegalArgumentException("Wrong parameters");
        }

        return init(host, port, cipherSuiteFilter, sniHostName);
    }

    static SSLClient init(String host, int port, String cipherSuiteFilter, String sniHostName) throws NoSuchAlgorithmException, IOException {
        SSLContext sslContext = SSLContext.getDefault();
        SSLSocketFactory ssf = (SSLSocketFactory) sslContext.getSocketFactory();
        SSLSocket socket = (SSLSocket) ssf.createSocket(host, port);

        if(cipherSuiteFilter != null) {
            String[] cipherSuites = Utils.filterStringArray(ssf.getSupportedCipherSuites(), cipherSuiteFilter);
            logger.info("Enabled cipher suites: " + Arrays.toString(cipherSuites));
            socket.setEnabledCipherSuites(cipherSuites);
        }

        if(sniHostName != null) {
            logger.info("Set SNI hostname: " + sniHostName);
            SNIHostName serverName = new SNIHostName(sniHostName);
            List<SNIServerName> serverNames = new ArrayList<>();
            serverNames.add(serverName);
            SSLParameters params = socket.getSSLParameters();
            params.setServerNames(serverNames);
            socket.setSSLParameters(params);
        }

        return new SSLClient(socket);
    }

    private SSLClient(SSLSocket socket) {
        this.socket = socket;
        dataExchanger = new ByteArrayDataExchanger();
    }

    void connect() throws IOException {
        dataExchanger.exchange(socket.getInputStream(), socket.getOutputStream());
        if(!dataExchanger.isSuccessful()) {
            throw new RuntimeException("TEST FAILED: data exchnage failed");
        }

        logger.log(Level.INFO, "TEST PASSED");
    }
}
