import javax.net.ssl.SSLSocket;
import javax.net.ssl.SSLServerSocket;
import javax.net.ssl.SSLServerSocketFactory;
import javax.net.ssl.SSLContext;
import java.io.IOException;
import java.io.File;
import java.security.NoSuchAlgorithmException;
import java.security.PrivilegedExceptionAction;
import java.util.ArrayList;
import java.util.logging.Logger;
import java.util.logging.Level;
import java.util.Arrays;
import java.util.List;
import javax.net.ssl.SNIHostName;
import javax.net.ssl.SNIMatcher;
import javax.net.ssl.SSLParameters;
import javax.security.auth.Subject;
import javax.security.auth.login.LoginContext;

class SSLEchoServer implements Runnable {

    private final static Logger logger = Logger.getLogger(SSLEchoServer.class.getName());

    private final static String STOP_FILE_NAME = "ssl_echo_server.run";

    private static File stopFile;

    private SSLServerSocket serverSocket;

    public static void main(final String[] args) throws Exception {
        String jaasCongifName = System.getProperty("jaasConfigName");
        if(jaasCongifName != null && !jaasCongifName.isEmpty()) {
            LoginContext sslEchoServerLoginContext = new LoginContext(jaasCongifName);
            sslEchoServerLoginContext.login();
            logger.info("SSL Echo server authentication successful");
            Subject sslEchoServerSubject = sslEchoServerLoginContext.getSubject();

            Subject.doAs(sslEchoServerSubject, new PrivilegedExceptionAction<Object> () {

                @Override
                public Object run() throws Exception {
                    startServer(args);
                    return null;
                }

            });
        }
        else {
            startServer(args);
        }
    }

    public static void startServer(String[] args) throws IOException, NoSuchAlgorithmException {
        SSLEchoServer sslEchoServer = SSLEchoServer.init(args);

        stopFile = new File(STOP_FILE_NAME);
        if(!stopFile.exists()) {
            boolean success = stopFile.createNewFile();
            if(!success) throw new IOException("Cannot create '" + STOP_FILE_NAME + "' file");
        }

        sslEchoServer.start();

        while(true) {
            if(isStopped()) break;
            try {
                Thread.sleep(5000);
            } catch(InterruptedException e) {
                // ignore
            }
        }
    }

    static private boolean isStopped() {
        return !stopFile.exists();
    }

    private static SSLEchoServer init(String[] args) throws NoSuchAlgorithmException, IOException {
        int port = 0;
        String cipherSuiteFilter = null;
        String sniMatcherPattern = null;

        try {
            int i = 0;
            String arg;
            while (i < args.length) {
                arg = args[i++];
                switch (arg) {
                    case "-port":
                        if (i < args.length)
                            port = Integer.parseInt(args[i++]);
                        else
                            throw new Exception("-port requires a port number");
                        break;
                    case "-cipher_suite_filter":
                        if (i < args.length)
                            cipherSuiteFilter = args[i++];
                        else
                            throw new Exception("-cipher_suite_filter requires a value");
                        break;
                    case "-sni_matcher_pattern":
                        if (i < args.length)
                            sniMatcherPattern = args[i++];
                        else
                            throw new Exception("-sni_matcher_pattern requires a value");
                        break;
                    default:
                        throw new Exception("Unrecognized option: " + arg);
                }
            }
        } catch(Exception e) {
            System.out.println("Parsing command line options failed: " + e.getMessage());
            System.out.println("Usage: " + SSLEchoServer.class.getName() + " -port 1234");
            throw new IllegalArgumentException("Wrong parameters");
        }

        if(port <=0 || port > 65535) {
            System.out.println("Wrong port number, or port is not set. Use -port parameter");
            throw new IllegalArgumentException("Wrong parameters");
        }

        return init(port, cipherSuiteFilter, sniMatcherPattern);
    }

    static SSLEchoServer init(int port, String cipherSuiteFilter, String sniMatcherPattern) throws NoSuchAlgorithmException, IOException {
        SSLContext sslContext = SSLContext.getDefault();
        SSLServerSocketFactory ssf = (SSLServerSocketFactory) sslContext.getServerSocketFactory();
        SSLServerSocket serverSocket = (SSLServerSocket) ssf.createServerSocket(port);

        if(cipherSuiteFilter != null) {
            String[] cipherSuites = Utils.filterStringArray(ssf.getSupportedCipherSuites(), cipherSuiteFilter);
            logger.info("Enabled cipher suites: " + Arrays.toString(cipherSuites));
            serverSocket.setEnabledCipherSuites(cipherSuites);
        }

        if(sniMatcherPattern != null) {
            logger.info("Set SNI matcher: " + sniMatcherPattern);
            SNIMatcher matcher = SNIHostName.createSNIMatcher(sniMatcherPattern);
            List<SNIMatcher> matchers = new ArrayList<>();
            matchers.add(matcher);
            SSLParameters params = serverSocket.getSSLParameters();
            params.setSNIMatchers(matchers);
            serverSocket.setSSLParameters(params);
        }

        return new SSLEchoServer(serverSocket);
    }

    private SSLEchoServer(SSLServerSocket serverSocket) {
        this.serverSocket = serverSocket;
    }

    void start() throws IOException {
        Thread serverThread = new Thread(this, "SSL echo server thread");
        serverThread.setDaemon(true);
        serverThread.start();
    }

    @Override
    public void run() {
        // main server loop
        logger.info("Start server loop");
        while (true) {
            try {
                SSLSocket clientSocket = (SSLSocket) serverSocket.accept();
                logger.info("Client connection accepted");
                if (clientSocket != null) {
                    // process client connection
                    Thread clientThread = new Thread(createConnectionHandler(clientSocket), "Client connection handler thread (port " + clientSocket.getPort() + ")");
                    clientThread.setDaemon(true);
                    clientThread.start();
                }
            } catch (IOException e) {
                logger.log(Level.SEVERE, "Cannot accept client connection", e);
            }
        }
    }

    /**
     * Creates a handler for client connection.
     *
     * @param socket Client socket
     * @return Handler for client connection
     */
    private Runnable createConnectionHandler(SSLSocket socket) {
        return new SSLEchoConnectionHandler(socket);
    }
}
